/* vrJuggler interface
 * Copyright (C) 2001 Paul Rajlich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <GL/gl.h>

// vrJuggler stuff
#include <Kernel/vjKernel.h>
#include <Kernel/GL/vjGlApp.h>
#include <Kernel/GL/vjGlContextData.h>
#include <Kernel/vjDebug.h>
#include <Kernel/vjProjection.h>
#include <Input/InputManager/vjPosInterface.h>
#include <Input/InputManager/vjDigitalInterface.h>
#include <Performance/vjTimeStamp.h>

// timing
#include <sys/time.h>
#include <unistd.h>

#include "vec.h"
#include "uicommon.h"
#include "rendercontext.h"
#include "globalshared.h"

global_shared_t *g;

void printUsage();

// use regular memory (juggler uses threads)
void *gc_malloc(size_t size) { return malloc(size); }
void *rc_malloc(size_t size) { return malloc(size); }
void gc_free(void *mem) { free(mem); }
void rc_free(void *mem) { free(mem); }

// timing
#ifdef WIN32
double gettime()
{
   return ((double)GetTickCount())/1000.0f;
}
#else
double gettime()
{
    struct timeval tv;
    gettimeofday(&tv, 0);
    return  ((double) tv.tv_sec + (double) tv.tv_usec / 1000000.0);
}
#endif


// Wrapper class to hold a copy of a context specific data structure
class ContextData
{
public:
   ContextData() : r_context(NULL)
   {
      r_context = (r_context_t *) rc_malloc(sizeof(r_context_t));
      std::cout << "Allocated ContextData." << std::endl;
   }

   r_context_t* getData()
   { return r_context; }

private:
   r_context_t* r_context;
};


// Actual application class
// Just derives from vjGlApp base class and implments the interface methods
// needed to call the rendering engine.
class myApp : public vjGlApp {
private:
  // see rendercontext.h. Each rendering thread will have its own copy
  vjGlContextData<ContextData> contextData;

public:
  vjPosInterface mWand;
  vjDigitalInterface mButton0;
  vjDigitalInterface mButton1;
  vjDigitalInterface mButton2;

  double starttime, frametime, lastTime;

public:
  myApp(vjKernel *kern) : vjGlApp(kern) {
    mWand.init("VJWand");
    mButton0.init("VJButton0");
    mButton1.init("VJButton1");
    mButton2.init("VJButton2");
    std::cerr << std::endl;
    std::cerr << "***************************************" << std::endl;
    std::cerr << "bsp: " << bsp << " pak: " << pak << std::endl;
    std::cerr << "***************************************" << std::endl;
    std::cerr << std::endl;
  }

   // Initialize global data structures
  virtual void init() {

    // load the map
    ui_init_bsp();
    printf("\n\n");

    starttime = gettime();
    frametime = starttime;
    lastTime = starttime;

    g->r_setup_projection = 0;
  }

  // Initialize context data
  virtual void contextInit() {
    r_context_t* the_context_data = contextData->getData();
    vjASSERT(the_context_data != NULL && "context data pointer should be non-NULL");

    ui_init_gl(the_context_data);
  }

  // Execute the actually drawing
  virtual void draw() {
    r_context_t* the_context_data = contextData->getData();
    vjASSERT(the_context_data != NULL);

    // Set some state that is expected
    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glCullFace(GL_FRONT);
    glShadeModel(GL_SMOOTH);

    ui_display(the_context_data, frametime - starttime);
    glDisable(GL_BLEND);

    // reset some state to make simulator mode look "better"
    glCullFace(GL_BACK);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glShadeModel(GL_FLAT);
  }

  // Updates that occur immedately after device updates but before rendering
  // - Update timing
  // - Update navigation
  virtual void preFrame() {
    double etime;
    float wandH(0.0), wandP(0.0), wandR(0.0);

    frametime = gettime();
    etime = frametime - lastTime;
    if (etime > 0.2f)
      etime = 0.2f; /* cap */

    // move forward
    if (mButton0->getData())
      ui_key_down('w');
    else
      ui_key_up('w');

    // change sky
    if (mButton2->getData() == vjDigital::TOGGLE_ON)
      ui_key_down('b');

    // turn on/off collision detection
    if (mButton1->getData() == vjDigital::TOGGLE_ON)
      ui_key_down('m');

    // travel in direction wand is pointing
    vjMatrix *wandMat = mWand->getData();
    wandH = wandMat->getYRot();
    wandP = wandMat->getXRot();

    vjMatrix rotated_wand;             // Wand position with correction for current rotation
    vjMatrix correction_mat;           // Matrix to account for current model/world rotation (eye_az)
    correction_mat.makeXYZEuler(0,g->r_eye_az, 0);
    rotated_wand.mult(correction_mat, *wandMat);

    vjVec3 forward_vec(0.0,0.0,-1.0f);    // Forward in vj world coords
    vjVec3 eye_dir;
    eye_dir.xformVec(rotated_wand, forward_vec);

    // Flip to engine coordinate system
    g->r_movedir[0] = -eye_dir[2];    // +z
    g->r_movedir[1] = -eye_dir[0];    // +x
    g->r_movedir[2] = eye_dir[1];     // -y

    // Rotate the world
    const float az_scale(1.0f);
    if ( (fabsf(wandP) < 75.0f) || ( fabsf(wandP) > 105.0f) )        // Set clipping region when pointing directly down/up
    {
       if (fabsf(wandH) > 40.0f)
       {
         if(wandH > 0.0f)
            g->r_eye_az += (wandH-40.0f)*etime*az_scale;
         else
            g->r_eye_az += (wandH+40.0f)*etime*az_scale;
       }
    }
    //else vjDEBUG(0,0) << "Clipping az because wandP is --> " << wandP << std::endl << vjDEBUG_FLUSH;

    ui_move(frametime - starttime);

    lastTime = frametime;
  }
};


int
main(int argc, char **argv)
{
  // setup global data structure
  g = (global_shared_t *) gc_malloc(sizeof(global_shared_t));
  init_global_shared(g);

  // USAGE:
  // command -c configfile1 -c configfile2 -- otherOptions
  // NOTE: The config files all have to be first then the "--" specification
  if(argc <= 1)
    printUsage();

  vjKernel* kernel = vjKernel::instance();

  // Get the Juggler commands first
  int c;
  while ( (c = getopt(argc, argv, "hc:")) != -1)
  {
      switch (c) {
      case 'c':
           kernel->loadConfigFile( std::string(optarg) );
           break;
      case 'h':
            printUsage();
            exit(1);
            break;
      case '?':
           break;
      }
  }

  // optind-1 because it will ignore elt 0 (normally app name)
  ui_read_args((argc-optind+1), &(argv[optind-1]));
  strcpy(g->r_help_fname, "paul/helpVR.jpg");

  myApp *app = new myApp(kernel);

  kernel->start();
  kernel->setApplication(app);

  while (1) {
    usleep(100000);
  }
}

void printUsage()
{
   std::cout << "Usage: cq3a -c <juggler-config-file> -c <more-juggler-config> -- <cq3a config params>" << std::endl;
}

